class Pyz3rException(Exception):
    pass


class AlttprFailedToRetrieve(Pyz3rException):
    pass


class AlttprFailedToGenerate(Pyz3rException):
    pass


class UnableToRetrieve(Pyz3rException):
    pass

class UnableToGenerate(Pyz3rException):
    pass