"""
ALTTPR API Wrapper
~~~~~~~~~~~~~~~~~~~
A basic wrapper for the ALTTPR.com API.
"""

__title__ = 'pyz3r'
__author__ = 'Thomas Prescott'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2018-2019 Thomas Prescott'
__version__ = '6.0.0'

from pyz3r.alttpr import ALTTPR
from pyz3r.sm import sm
from pyz3r.smvaria import SuperMetroidVaria
from pyz3r import mystery, customizer, rom